import torch
import torch.nn as nn
import torch.nn.functional as F
from tests.internal.common_utils import find_free_port
from tests.internal.compressor import MinMaxUInt8
import unittest
import torch.multiprocessing as mp
import os
from bagua.torch_api.utils import apply_flattened_call, flatten
import bagua.torch_api as bagua


class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.fc1 = nn.Linear(2, 10, bias=False)
        self.fc2 = nn.Linear(10, 50, bias=True)
        self.fc3 = nn.Linear(50, 4, bias=False)
        self.relu = nn.ReLU()

    def forward(self, x):
        x = self.relu(self.fc1(x))
        x = self.relu(self.fc2(x))
        x = self.fc3(x)
        return F.softmax(x, dim=1)


def _init_env(rank):
    # set deterministic
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True
    torch.manual_seed(rank)
    # initialize subprocess env
    os.environ["RANK"] = str(rank)
    os.environ["LOCAL_RANK"] = str(rank)


def run_model(rank, nprocs, hierarchical, communication_interval, results):
    _init_env(rank)

    # init bagua distributed process group
    torch.cuda.set_device(rank)
    bagua.init_process_group()

    # construct model and optimizer, etc.
    model = Net().cuda()
    optimizer = torch.optim.SGD(model.parameters(), lr=0.01)
    loss_fn = nn.MSELoss()

    # wrap model
    model = model.with_bagua(
        [optimizer],
        bagua.algorithms.decentralized.LowPrecisionDecentralizedAlgorithm(
            hierarchical=hierarchical,
            communication_interval=communication_interval,
        ),
    )

    ret = results[rank]
    bucket = model.bagua_buckets[0]

    ret.init_weight.copy_(flatten([param.data for param in model.parameters()]))

    for epoch in range(10):
        data = torch.randn(4, 2).cuda()
        target = torch.randn(4, 4).cuda()

        optimizer.zero_grad()
        output = model(data)
        loss = loss_fn(output, target)

        loss.backward()
        optimizer.step()

    ret.bucket_weight.copy_(flatten([param.data for param in model.parameters()]))
    ret.weight.copy_(torch.norm(bucket._weight))
    ret.left_peer_weight.copy_(torch.norm(bucket._left_peer_weight))
    ret.right_peer_weight.copy_(torch.norm(bucket._right_peer_weight))


def run_torch_model(
    rank, nprocs, hierarchical, communication_interval, results, backend
):
    _init_env(rank)

    # init torch distributed process group
    torch.cuda.set_device(rank)
    store = torch.distributed.FileStore("/tmp/filestore", nprocs)
    torch.distributed.init_process_group(
        world_size=nprocs, rank=rank, store=store, backend=backend
    )

    # construct model and optimizer, etc.
    model = Net().cuda()
    optimizer = torch.optim.SGD(model.parameters(), lr=0.01)
    loss_fn = nn.MSELoss()

    # wrap model
    model = LowPrecDecentralizedAlgor(
        model, optimizer, hierarchical, communication_interval
    )

    ret = results[rank]
    ret.init_weight.copy_(flatten([param.data for param in model.parameters()]))

    for epoch in range(10):
        data = torch.randn(4, 2).cuda()
        target = torch.randn(4, 4).cuda()

        optimizer.zero_grad()
        output = model(data)
        loss = loss_fn(output, target)

        loss.backward()
        model.step()

    ret.bucket_weight.copy_(flatten([param.data for param in model.parameters()]))


class Result(object):
    def __init__(self):
        model = Net()
        self.init_weight = flatten(
            [torch.zeros_like(param.data) for param in model.parameters()]
        )
        self.bucket_weight = flatten(
            [torch.zeros_like(param.data) for param in model.parameters()]
        )
        self.weight = torch.Tensor([0.0])
        self.left_peer_weight = torch.Tensor([0.0])
        self.right_peer_weight = torch.Tensor([0.0])


class LowPrecDecentralizedAlgor(nn.Module):
    def __init__(self, module, optimizer, hierarchical, communication_interval):
        super(LowPrecDecentralizedAlgor, self).__init__()
        self.module = module
        self.optimizer = optimizer
        self.hierarchical = hierarchical
        self.communication_interval = communication_interval
        self.compressor = MinMaxUInt8()
        self.step_count = 0

        assert torch.distributed.is_initialized()

        self.rank = torch.distributed.get_rank()
        self.world_size = torch.distributed.get_world_size()

        # broadcast parameters
        for param in self.module.parameters():
            torch.distributed.broadcast(param.data, src=0)

        self.weight = flatten(self._build_params()).cuda()
        self.left_peer_weight = self.weight.detach().clone().cuda()
        self.right_peer_weight = self.weight.detach().clone().cuda()

    def _build_params(self):
        return [param.data for param in list(self.module.parameters()).__reversed__()]

    def _should_communicate(self):
        return self.step_count % self.communication_interval == 0

    def forward(self, *inputs, **kwargs):
        result = self.module(*inputs, **kwargs)
        return result

    def step(self):
        self.optimizer.step()

        def communicate_with_peers(
            tensor: torch.Tensor, comm_size: int
        ) -> (torch.Tensor, torch.Tensor):
            if comm_size == 1:
                return tensor, tensor

            tensor = tensor.cpu()
            left_tensor = torch.zeros_like(tensor)
            right_tensor = torch.zeros_like(tensor)

            left_peer_rank = (self.rank + self.world_size - 1) % comm_size
            right_peer_rank = (self.rank + 1) % comm_size

            requests = []
            requests.append(torch.distributed.isend(tensor, left_peer_rank))
            requests.append(torch.distributed.isend(tensor, right_peer_rank))
            requests.append(torch.distributed.irecv(left_tensor, left_peer_rank))
            requests.append(torch.distributed.irecv(right_tensor, right_peer_rank))

            for req in requests:
                req.wait()

            return left_tensor.cuda(), right_tensor.cuda()

        def update_weight_fn(x, comm_size):
            x += 1 / 3 * self.left_peer_weight
            x += 1 / 3 * self.right_peer_weight
            x -= 5 / 3 * self.weight

            minmax, compressed = self.compressor.compress(x)
            left_compressed, right_compressed = communicate_with_peers(
                compressed, comm_size
            )
            left_minmax, right_minmax = communicate_with_peers(minmax, comm_size)

            self.left_peer_weight += self.compressor.decompress(
                left_minmax, left_compressed
            )
            self.right_peer_weight += self.compressor.decompress(
                right_minmax, right_compressed
            )

            diff = self.compressor.decompress(minmax, compressed)
            x.copy_(self.weight + diff)
            self.weight.copy_(x)

        def hierarchical_update_weight_fn(x):
            torch.distributed.reduce(x, dst=0)
            if self.rank == 0:
                x /= self.world_size
                update_weight_fn(x, comm_size=1)

            torch.distributed.broadcast(x, 0)

        if self._should_communicate():
            weights = self._build_params()
            if self.hierarchical:
                apply_flattened_call(weights, hierarchical_update_weight_fn)
            else:
                apply_flattened_call(
                    weights, lambda x: update_weight_fn(x, self.world_size)
                )
        self.step_count += 1


class TestLowPrecisionDecentralized(unittest.TestCase):
    def run_test_locally(self, hierarchical, communication_interval):
        if not torch.cuda.is_available():
            print("skip tests since cuda is not available")
            return

        nprocs = torch.cuda.device_count()
        os.environ["WORLD_SIZE"] = str(nprocs)
        os.environ["LOCAL_WORLD_SIZE"] = str(nprocs)
        os.environ["MASTER_ADDR"] = "127.0.0.1"
        os.environ["MASTER_PORT"] = str(find_free_port())
        os.environ["BAGUA_SERVICE_PORT"] = str(find_free_port())

        results = [Result() for _ in range(nprocs)]
        mp.spawn(
            run_model,
            nprocs=nprocs,
            args=(nprocs, hierarchical, communication_interval, results),
        )

        for rank in range(nprocs):
            left_peer_rank = (rank + nprocs - 1) % nprocs
            right_peer_rank = (rank + 1) % nprocs

            if hierarchical:
                # all workers have equal weights
                self.assertTrue(
                    torch.equal(
                        results[rank].bucket_weight,
                        results[left_peer_rank].bucket_weight,
                    )
                )
            else:
                self.assertTrue(
                    torch.equal(
                        results[rank].weight, results[left_peer_rank].right_peer_weight
                    )
                )
                self.assertTrue(
                    torch.equal(
                        results[rank].weight, results[right_peer_rank].left_peer_weight
                    )
                )

    def run_diff_locally(self, hierarchical, communication_interval, backend):
        if not torch.cuda.is_available():
            print("skip tests since cuda is not available")
            return

        nprocs = torch.cuda.device_count()
        os.environ["WORLD_SIZE"] = str(nprocs)
        os.environ["LOCAL_WORLD_SIZE"] = str(nprocs)
        os.environ["MASTER_ADDR"] = "127.0.0.1"
        os.environ["MASTER_PORT"] = str(find_free_port())
        os.environ["BAGUA_SERVICE_PORT"] = str(find_free_port())

        torch_results = [Result() for _ in range(nprocs)]
        mp.spawn(
            run_torch_model,
            nprocs=nprocs,
            args=(nprocs, hierarchical, communication_interval, torch_results, backend),
        )

        bagua_results = [Result() for _ in range(nprocs)]
        mp.spawn(
            run_model,
            nprocs=nprocs,
            args=(nprocs, hierarchical, communication_interval, bagua_results),
        )

        for rank in range(nprocs):
            self.assertTrue(
                torch.all(
                    torch.isclose(
                        bagua_results[rank].init_weight,
                        torch_results[rank].init_weight,
                    )
                ).item()
            )

            self.assertTrue(
                torch.all(
                    torch.isclose(
                        bagua_results[rank].bucket_weight,
                        torch_results[rank].bucket_weight,
                    )
                ).item()
            )

    def test_algorithm(self):
        self.run_test_locally(hierarchical=False, communication_interval=1)
        self.run_test_locally(hierarchical=False, communication_interval=2)
        self.run_test_locally(hierarchical=True, communication_interval=1)

    def test_compare(self):
        self.run_diff_locally(
            hierarchical=False, communication_interval=1, backend="gloo"
        )
        self.run_diff_locally(
            hierarchical=False, communication_interval=2, backend="gloo"
        )
        self.run_diff_locally(
            hierarchical=True, communication_interval=1, backend="nccl"
        )


if __name__ == "__main__":
    unittest.main()



from collections import OrderedDict
import torch.distributed as dist
import torch
import math
import time
import logging
import numpy as np
from typing import Tuple, List

LOGGER = logging.getLogger(__name__)

flatten = torch._utils._flatten_dense_tensors
unflatten = torch._utils._unflatten_dense_tensors


def apply_flattened_call(bucket, call, extra_args=None):
    coalesced = flatten(bucket)

    if extra_args is not None:
        call(coalesced, *extra_args)
    else:
        call(coalesced)

    if call is dist.all_reduce:
        coalesced /= dist.get_world_size()

    for buf, synced in zip(bucket, unflatten(coalesced, bucket)):
        buf.copy_(synced)


def _group_by_tensor_type(tensors):
    buckets = OrderedDict()
    for tensor in tensors:
        tp = tensor.type()
        if tp not in buckets:
            buckets[tp] = []
        buckets[tp].append(tensor)
    return buckets


def apply_flattened_call_all(tensors, call):
    """
    Apply call on a list of tensors.
    """

    grouped_tensors = _group_by_tensor_type(tensors)
    for tensors in grouped_tensors.values():
        apply_flattened_call(tensors, call)


def align_size(size, align):
    return int((size + align - 1) / align) * align


def check_contiguous(tensors):
    data_ptr = None
    for t in tensors:
        if data_ptr is not None and t.data_ptr() != data_ptr:
            return False
        data_ptr = t.data_ptr() + t.numel() * t.element_size()
    return True


def _get_params_flattened_aligned_size(params, align_bytes):
    assert align_bytes == 1 or (
        align_bytes % params[0].element_size() == 0
    ), "align bytes must be multiples of element size"

    sizes = [p.numel() for p in params]

    total_size = sum(sizes)
    aligned_total_size = (
        align_size(total_size * params[0].element_size(), align_bytes)
        // params[0].element_size()
    )

    # padding to the last param
    sizes[-1] += aligned_total_size - total_size

    for p, sz in zip(params, sizes):
        p.allocated_size = sz

    return aligned_total_size


def flatten_module_params(params_list, align_bytes: int):
    if len(params_list) == 0:
        return

    if not isinstance(params_list[0], list):
        params_list = [params_list]

    total_size = 0
    for params in params_list:
        total_size += _get_params_flattened_aligned_size(params, align_bytes)

    logging.debug(
        f"flatten {str(params_list[0][0].dtype).partition('.')[-1]} params aligned to {align_bytes} bytes, total numels: {total_size}"
    )

    flatten_weights_tensor = torch.zeros(total_size, dtype=params_list[0][0].dtype).to(
        params_list[0][0].device
    )
    flatten_grads_tensor = torch.zeros(total_size, dtype=params_list[0][0].dtype).to(
        params_list[0][0].device
    )

    flatten_weights_storage = flatten_weights_tensor.storage()
    flatten_grads_storage = flatten_grads_tensor.storage()

    def set_storage(param, weight_storage, grad_storage, storage_offset):
        with torch.no_grad():
            z = torch.zeros_like(param.data)
            z.set_(weight_storage, storage_offset, param.shape)
            param.data = z

            t = torch.zeros_like(param.data)
            t.set_(grad_storage, storage_offset, param.shape)
            param.grad = t

    offset = 0
    for params in params_list:
        for p in params:
            # copy data
            flatten_weights_tensor[offset : offset + p.numel()] = p.data.reshape(-1)

            if p.grad is not None:
                flatten_grads_tensor[offset : offset + p.numel()] = p.grad.data.reshape(
                    -1
                )
            else:
                logging.debug(f"grad is none, {offset}")

            # flatten
            set_storage(p, flatten_weights_storage, flatten_grads_storage, offset)
            offset += p.allocated_size
            logging.debug(f"flatten param done {offset}")

    # # check
    for params in params_list:
        weight_tensors = [p.data for p in params]
        grad_tensors = [p.grad.data for p in params]

        assert check_contiguous(weight_tensors)
        assert check_contiguous(grad_tensors)

    return new_param(flatten_weights_tensor, flatten_grads_tensor)


def collocate_params(params):
    """
    `tensors` share the same storage
    """
    if len(params) == 1:
        return params[0]

    logging.debug(f"fuse {len(params)} params")

    sorted_params = sorted(params, key=lambda x: x.storage_offset())

    start = None
    offset = 0
    for p in sorted_params:
        weight = p.data
        grad = p.grad.data

        assert (
            weight.storage_offset() == grad.storage_offset()
        ), "collocated weights and grads must have consistent storage offset"

        if start is None:
            start = offset = weight.storage_offset()
        else:
            assert (
                offset == weight.storage_offset()
            ), "params collocated must be contiguous"

        offset += (
            p.bagua_tensor.num_elem_allocated()
            if hasattr(p, "bagua_tensor")
            else p.numel()
        )

    with torch.no_grad():
        weight_tensor = torch.zeros(offset - start, dtype=params[0].dtype).to(
            params[0].device
        )
        weight_tensor.set_(params[0].data.storage(), start, weight_tensor.shape)

        grad_tensor = torch.zeros(offset - start, dtype=params[0].dtype).to(
            params[0].device
        )
        grad_tensor.set_(params[0].grad.data.storage(), start, grad_tensor.shape)

        return new_param(weight_tensor, grad_tensor)


def new_param(weight, grad):
    with torch.no_grad():
        p = torch.nn.Parameter(weight, requires_grad=False)
        p.grad = grad
        return p


def to_bagua_datatype(datatype):
    if datatype == torch.float32:
        return "f32"
    elif datatype == torch.float16:
        return "f16"
    elif datatype == torch.uint8:
        return "u8"
    elif datatype == torch.long:
        return "i64"
    else:
        raise ValueError(f"unsupported data type {datatype}.")


def average_by_removing_extreme_values(raw_score_list):
    score_list = np.asarray(raw_score_list)

    # weed out warm up data
    score_list = score_list[len(score_list) // 3 :]

    def weed_out_outliers(X):
        mean = np.mean(X)
        std = np.std(X)
        distance_from_mean = abs(X - mean)
        max_deivations = 1
        not_outlier = distance_from_mean < max_deivations * std
        not_outliers = X[not_outlier]

        if len(not_outliers) == 0:
            return X

        return not_outliers

    score_list = weed_out_outliers(score_list)

    # Repeat up to ten times
    for i in range(10):
        if np.std(score_list) < np.mean(score_list):
            break
        score_list = weed_out_outliers(score_list)

    # score = np.mean(score_list) # TODO: @shjwudp check whether these are still needed
    # std = np.std(score_list)

    return np.mean(score_list), np.std(score_list), score_list.tolist()


class StatisticalAverage:
    def __init__(
        self,
        last_update_time: float = time.time(),
        records: List[float] = [],
        record_tail: Tuple[float, float] = (0.0, 0.0),  # [tail_len, tail_val]
    ) -> None:
        """Track and record the average over a period of time.

        Args:
            last_update_time (float, optional): last update time.
                Defaults to time.time().
            records (List[float], optional): statistical average value from
                `last_update_time`, records[i] is the average value from
                last_update_time to last_update_time + 2 ^ i (unit: seconds).
                Defaults to [].
            tail (Tuple[float, float], optional): tail of record, first one
                is tail length (unit: seconds), second one is tail average
                value. Defaults to (0., 0.).
        """
        self.last_update_time: float = last_update_time
        self.records: List[float] = records
        self.record_tail: Tuple[float, float] = record_tail

    def record_seconds(self) -> float:
        return 2.0 ** (len(self.records) - 1) if len(self.records) != 0 else 0.0

    def total_recording_time(self) -> float:
        (tail_seconds, _) = self.record_tail

        return self.record_seconds() + tail_seconds

    def get_records_mean(self, last_n_seconds: float) -> float:
        if last_n_seconds <= 0.0:
            return 0.0

        records_seconds = self.record_seconds()
        (tail_seconds, tail_mean) = self.record_tail

        if len(self.records) == 0:
            return tail_mean

        if last_n_seconds < 1.0:
            return self.records[0]

        if last_n_seconds <= records_seconds:
            floor_id = max(0, math.floor(math.log(last_n_seconds, 2.0)))
            floor_time = 2.0 ** floor_id
            if floor_id + 1 < len(self.records):
                a, b = self.records[floor_id], self.records[floor_id + 1]
                a_l, b_l = floor_time, floor_time * 2.0
                mean = a + (b - a) * (last_n_seconds - a_l) / (b_l - a_l)
            else:
                mean = self.records[floor_id]
        elif last_n_seconds <= records_seconds + tail_seconds:
            a, b = self.records[-1], tail_mean
            a_l, b_l = records_seconds, records_seconds + tail_seconds
            mean = a + (b - a) * (last_n_seconds - a_l) / (b_l - a_l)
        else:
            mean = tail_mean

        return mean

    def record(self, val: float):
        now = time.time()
        time_dist: float = now - self.last_update_time
        new_records: List[float] = []
        new_tail: Tuple[float, float] = (0.0, 0.0)

        for i in range(64):
            coverage_period = 2.0 ** i

            if coverage_period <= time_dist:
                new_records.append(val)
            elif coverage_period <= time_dist + self.total_recording_time():
                record_contribution_percentage = time_dist / coverage_period
                new_val = val * record_contribution_percentage + self.get_records_mean(
                    coverage_period - time_dist
                ) * (1.0 - record_contribution_percentage)

                new_records.append(new_val)

                if coverage_period > time_dist + self.total_recording_time():
                    break
            else:
                new_total_time = time_dist + self.total_recording_time()
                report_contribution_percentage = time_dist / new_total_time

                tail_len = new_total_time - 2.0 ** (len(new_records) - 1)
                tail_val = val * report_contribution_percentage + self.get_records_mean(
                    self.total_recording_time()
                ) * (1.0 - report_contribution_percentage)
                new_tail = (tail_len, tail_val)
                break

        self.last_update_time = now
        self.records = new_records
        self.record_tail = new_tail

    def get(self, last_n_seconds: float) -> float:
        time_dist = time.time() - self.last_update_time
        if last_n_seconds <= time_dist:
            if len(self.records) != 0:
                return self.records[0]
            else:
                (tail_mean, _) = self.record_tail
                return tail_mean

        return self.get_records_mean(last_n_seconds - time_dist)

    def __str__(self) -> str:
        return str(
            {
                "last_update_time": self.last_update_time,
                "records": self.records,
                "record_tail": self.record_tail,
            }
        )



import torch
import torch.nn as nn
import torch.nn.functional as F
from tests.internal.common_utils import find_free_port
import unittest
import torch.multiprocessing as mp
import os
from bagua.torch_api.utils import flatten, unflatten
import bagua.torch_api as bagua


N_EPOCHS = 10


class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.fc1 = nn.Linear(2, 10, bias=False)
        self.fc2 = nn.Linear(10, 50, bias=True)
        self.fc3 = nn.Linear(50, 4, bias=False)
        self.relu = nn.ReLU()

    def forward(self, x):
        x = self.relu(self.fc1(x))
        x = self.relu(self.fc2(x))
        x = self.fc3(x)
        return F.softmax(x, dim=1)


def _init_env(rank):
    # set deterministic
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True
    torch.manual_seed(rank)
    # initialize subprocess env
    os.environ["RANK"] = str(rank)
    os.environ["LOCAL_RANK"] = str(rank)


def run_model(
    rank, nprocs, hierarchical, peer_selection_mode, communication_interval, results
):
    _init_env(rank)

    # init bagua distributed process group
    torch.cuda.set_device(rank)
    bagua.init_process_group()

    # construct model and optimizer, etc.
    model = Net().cuda()
    optimizer = torch.optim.SGD(model.parameters(), lr=0.01)
    loss_fn = nn.MSELoss()

    # wrap model
    model = model.with_bagua(
        [optimizer],
        bagua.algorithms.decentralized.DecentralizedAlgorithm(
            hierarchical=hierarchical,
            peer_selection_mode=peer_selection_mode,
            communication_interval=communication_interval,
        ),
    )

    ret = results[rank]

    ret.init_weight.copy_(flatten([param.data for param in model.parameters()]))

    for epoch in range(N_EPOCHS):
        data = torch.randn(4, 2).cuda()
        target = torch.randn(4, 4).cuda()

        optimizer.zero_grad()
        output = model(data)
        loss = loss_fn(output, target)

        loss.backward()
        optimizer.step()

    ret.bucket_weight.copy_(model.bagua_buckets[0]._peer_weight)


def run_torch_model(
    rank,
    nprocs,
    hierarchical,
    peer_selection_mode,
    communication_interval,
    results,
    backend,
):
    _init_env(rank)

    # init torch distributed process group
    torch.cuda.set_device(rank)
    store = torch.distributed.FileStore("/tmp/filestore", nprocs)
    torch.distributed.init_process_group(
        world_size=nprocs, rank=rank, store=store, backend=backend
    )

    # construct model and optimizer, etc.
    model = Net().cuda()
    optimizer = torch.optim.SGD(model.parameters(), lr=0.01)
    loss_fn = nn.MSELoss()

    # wrap model
    model = DecentralizedAlgor(
        model, optimizer, hierarchical, peer_selection_mode, communication_interval
    )

    ret = results[rank]
    ret.init_weight.copy_(flatten([param.data for param in model.parameters()]))

    for epoch in range(N_EPOCHS):
        data = torch.randn(4, 2).cuda()
        target = torch.randn(4, 4).cuda()

        optimizer.zero_grad()
        output = model(data)
        loss = loss_fn(output, target)

        loss.backward()
        model.step()

    ret.bucket_weight.copy_(model.peer_weight)


class Result(object):
    def __init__(self):
        model = Net()
        self.init_weight = flatten(
            [torch.zeros_like(param.data) for param in model.parameters()]
        )
        self.bucket_weight = flatten(
            [torch.zeros_like(param.data) for param in model.parameters()]
        )


class DecentralizedAlgor(nn.Module):
    def __init__(
        self,
        module,
        optimizer,
        hierarchical,
        peer_selection_mode,
        communication_interval,
    ):
        super(DecentralizedAlgor, self).__init__()
        self.module = module
        self.optimizer = optimizer
        self.hierarchical = hierarchical
        self.peer_selection_mode = peer_selection_mode
        self.communication_interval = communication_interval
        self.step_count = 0

        assert torch.distributed.is_initialized()

        self.rank = torch.distributed.get_rank()
        self.world_size = torch.distributed.get_world_size()

        # broadcast parameters
        for param in self.module.parameters():
            torch.distributed.broadcast(param.data, src=0)

    def _build_params(self):
        return [param.data for param in list(self.module.parameters()).__reversed__()]

    def communicate_with_peer(self):
        if self.peer_selection_mode == "all":
            torch.distributed.all_reduce(self.peer_weight)
            self.peer_weight /= self.world_size
        elif self.peer_selection_mode == "shift_one":
            peer_rank = get_peer_rank(
                self.peer_selection_mode,
                self.rank,
                self.world_size,
                self.step_count,
                self.communication_interval,
            )

            weight = self.weight.cpu()
            peer_weight = self.peer_weight.cpu()

            requests = []
            requests.append(torch.distributed.isend(weight, peer_rank))
            requests.append(torch.distributed.irecv(peer_weight, peer_rank))

            for req in requests:
                req.wait()

            self.peer_weight = peer_weight.cuda()
            self.weight = weight.cuda()

            self.peer_weight += self.weight
            self.peer_weight /= 2
        else:
            raise ValueError("Unsupported `peer_selection_mode`")

    def _should_communicate(self):
        return self.step_count % self.communication_interval == 0

    def forward(self, *inputs, **kwargs):
        if self._should_communicate():
            self.weight = flatten(self._build_params())
            self.peer_weight = flatten(self._build_params())
            self.communicate_with_peer()

        result = self.module(*inputs, **kwargs)
        return result

    def step(self):
        if self._should_communicate():
            params = self._build_params()
            for buf, synced in zip(params, unflatten(self.peer_weight, params)):
                buf.copy_(synced)

        self.optimizer.step()
        self.step_count += 1


def get_peer_rank(peer_selection_mode, rank, nranks, step, communication_interval):
    comm_step = step // communication_interval
    if peer_selection_mode == "shift_one":
        if rank < nranks // 2:
            return ((comm_step + rank) % ((nranks + 1) // 2)) + (nranks // 2)
        else:
            return (rank - (nranks // 2) - comm_step) % (nranks // 2)
    else:
        ValueError("Unsupported `peer_selection_mode`")


class TestLowPrecisionDecentralized(unittest.TestCase):
    def run_test_locally(
        self, nprocs, hierarchical, peer_selection_mode, communication_interval
    ):
        if not torch.cuda.is_available():
            print("skip tests since cuda is not available")
            return

        nprocs = torch.cuda.device_count()
        os.environ["WORLD_SIZE"] = str(nprocs)
        os.environ["LOCAL_WORLD_SIZE"] = str(nprocs)
        os.environ["MASTER_ADDR"] = "127.0.0.1"
        os.environ["MASTER_PORT"] = str(find_free_port())
        os.environ["BAGUA_SERVICE_PORT"] = str(find_free_port())

        results = [Result() for _ in range(nprocs)]
        mp.spawn(
            run_model,
            nprocs=nprocs,
            args=(
                nprocs,
                hierarchical,
                peer_selection_mode,
                communication_interval,
                results,
            ),
        )

        for rank in range(nprocs):
            if peer_selection_mode == "all":
                peer_rank = (rank + 1) % nprocs
                # all workers have equal weights
                self.assertTrue(
                    torch.equal(
                        results[rank].bucket_weight,
                        results[peer_rank].bucket_weight,
                    )
                )
            elif peer_selection_mode == "shift_one":
                peer_rank = get_peer_rank(
                    peer_selection_mode,
                    rank,
                    nprocs,
                    N_EPOCHS - 1,
                    communication_interval,
                )

                self.assertTrue(
                    torch.equal(
                        results[rank].bucket_weight, results[peer_rank].bucket_weight
                    )
                )
            else:
                raise ValueError("illegal `peer_selection_mode`!")

    def run_diff_locally(
        self, nprocs, hierarchical, peer_selection_mode, communication_interval, backend
    ):
        if not torch.cuda.is_available():
            print("skip tests since cuda is not available")
            return

        os.environ["WORLD_SIZE"] = str(nprocs)
        os.environ["LOCAL_WORLD_SIZE"] = str(nprocs)
        os.environ["MASTER_ADDR"] = "127.0.0.1"
        os.environ["MASTER_PORT"] = str(find_free_port())
        os.environ["BAGUA_SERVICE_PORT"] = str(find_free_port())

        torch_results = [Result() for _ in range(nprocs)]
        mp.spawn(
            run_torch_model,
            nprocs=nprocs,
            args=(
                nprocs,
                hierarchical,
                peer_selection_mode,
                communication_interval,
                torch_results,
                backend,
            ),
        )

        bagua_results = [Result() for _ in range(nprocs)]
        mp.spawn(
            run_model,
            nprocs=nprocs,
            args=(
                nprocs,
                hierarchical,
                peer_selection_mode,
                communication_interval,
                bagua_results,
            ),
        )

        for rank in range(nprocs):
            self.assertTrue(
                torch.all(
                    torch.isclose(
                        bagua_results[rank].init_weight,
                        torch_results[rank].init_weight,
                    )
                ).item()
            )

            self.assertTrue(
                torch.all(
                    torch.isclose(
                        bagua_results[rank].bucket_weight,
                        torch_results[rank].bucket_weight,
                    )
                ).item()
            )

    def test_algorithm(self):
        self.run_test_locally(
            nprocs=8,
            hierarchical=False,
            peer_selection_mode="all",
            communication_interval=1,
        )
        self.run_test_locally(
            nprocs=8,
            hierarchical=False,
            peer_selection_mode="shift_one",
            communication_interval=1,
        )
        self.run_test_locally(
            nprocs=8,
            hierarchical=False,
            peer_selection_mode="shift_one",
            communication_interval=2,
        )

    def test_compare(self):
        self.run_diff_locally(
            nprocs=8,
            hierarchical=False,
            peer_selection_mode="all",
            communication_interval=1,
            backend="gloo",
        )
        self.run_diff_locally(
            nprocs=8,
            hierarchical=False,
            peer_selection_mode="shift_one",
            communication_interval=1,
            backend="gloo",
        )
        self.run_diff_locally(
            nprocs=8,
            hierarchical=False,
            peer_selection_mode="shift_one",
            communication_interval=2,
            backend="gloo",
        )


if __name__ == "__main__":
    unittest.main()



import unittest
import logging
import multiprocessing
import socket
import time
import os
from flask import Flask
from typing import List
from bagua.bagua_define import TensorDeclaration, BaguaCoreTelemetrySpan
from bagua.service import AutotuneService, AutotuneClient
from bagua.bagua_define import BaguaHyperparameter, get_tensor_declaration_bytes


def pick_n_free_ports(n: int):
    socks = []
    for i in range(n):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.bind(("localhost", 0))
        socks.append(sock)

    n_free_ports = [sock.getsockname()[1] for sock in socks]
    for sock in socks:
        sock.close()

    return n_free_ports


def metrics(buckets, is_hierarchical_reduce):
    # score = sum(-abs(bucket_sum_size - 20MB))
    # convex function with peak at bucket_size=20MB
    score = 0.0
    for bucket in buckets:
        score += -abs(sum([get_tensor_declaration_bytes(td)
                      for td in bucket]) - 20 * 1024 ** 2)

    if not is_hierarchical_reduce:
        score += abs(score) * 0.1

    return score


class MockBaguaProcess:
    def __init__(
        self,
        rank: int,
        service_addr: str,
        service_port: int,
        model_name: str,
        tensor_list: List[TensorDeclaration],
        spans: List[BaguaCoreTelemetrySpan] = [],
    ) -> None:
        self.rank = rank
        self.model_name = model_name
        self.tensor_list = tensor_list
        self.spans = spans
        self.client = AutotuneClient(service_addr, service_port)

    def run(self):
        rsp = self.client.register_tensors(self.model_name, self.tensor_list)
        assert rsp.status_code == 200, "register_tensors failed, rsp={}".format(
            rsp)
        hp = BaguaHyperparameter().update(
            rsp.json()["recommended_hyperparameters"])

        train_iter = 0
        while True:
            score = metrics(hp.buckets, hp.is_hierarchical_reduce)
            rsp = self.client.report_metrics(
                self.model_name, self.rank, train_iter, hp.dict(), score
            )
            assert rsp.status_code == 200, "report_metrics failed, rsp={}".format(rsp)
            rsp = self.client.report_tensor_execution_order(self.spans)
            assert (
                rsp.status_code == 200
            ), "report_tensor_execution_order failed, rsp={}".format(rsp)
            rsp = self.client.ask_hyperparameters(
                self.model_name, self.rank, train_iter
            )
            assert rsp.status_code == 200, "ask_hyperparameters failed, rsp={}".format(
                rsp
            )
            rsp = rsp.json()
            hp.update(rsp["recommended_hyperparameters"])

            if rsp["is_autotune_completed"]:
                logging.info("train_iter={}".format(train_iter))
                break

            time.sleep(0.1)
            train_iter += 1

        return hp


class TestAutotuneService(unittest.TestCase):
    def test_autotune_service(self):
        service_addr = "127.0.0.1"
        service_port = pick_n_free_ports(1)[0]
        nprocs = 2

        autotune_service = AutotuneService(
            nprocs,
            autotune_level=1,
            sampling_confidence_time_s=0.1,
            warmup_time_s=1.0,
            is_output_autotune_log=True,
        )
        app = Flask(__name__)
        app = autotune_service.setup_app(app)
        log = logging.getLogger("werkzeug")
        log.setLevel(logging.ERROR)

        server = multiprocessing.Process(
            target=app.run,
            kwargs={
                "host": "0.0.0.0",
                "port": service_port,
                "debug": False,
            },
        )
        server.daemon = True
        server.start()

        model_dict = {
            "basic": ([
                TensorDeclaration(
                    {
                        "name": "basic.A",
                        "num_elements": 1 * 1024 ** 2,
                        "dtype": "f32",
                    }
                ),
                TensorDeclaration(
                    {
                        "name": "basic.B",
                        "num_elements": 2 * 1024 ** 2,
                        "dtype": "f32",
                    }
                ),
                TensorDeclaration(
                    {
                        "name": "basic.C",
                        "num_elements": 3 * 1024 ** 2,
                        "dtype": "f32",
                    }
                ),
                TensorDeclaration(
                    {
                        "name": "basic.D",
                        "num_elements": 4 * 1024 ** 2,
                        "dtype": "f32",
                    }
                ),
                TensorDeclaration(
                    {
                        "name": "basic.E",
                        "num_elements": 5 * 1024 ** 2,
                        "dtype": "f32",
                    }
                ),
            ], []),
            "Mixed_precision_test": ([
                TensorDeclaration(
                    {
                        "name": "Mixed_precision_test.A",
                        "num_elements": 1 * 1024 ** 2,
                        "dtype": "f32",
                    }
                ),
                TensorDeclaration(
                    {
                        "name": "Mixed_precision_test.B",
                        "num_elements": 3 * 1024 ** 2,
                        "dtype": "f32",
                    }
                ),
                TensorDeclaration(
                    {
                        "name": "Mixed_precision_test.C",
                        "num_elements": 5 * 1024 ** 2,
                        "dtype": "f16",
                    }
                ),
                TensorDeclaration(
                    {
                        "name": "Mixed_precision_test.D",
                        "num_elements": 7 * 1024 ** 2,
                        "dtype": "f16",
                    }
                ),
                TensorDeclaration(
                    {
                        "name": "Mixed_precision_test.E",
                        "num_elements": 11 * 1024 ** 2,
                        "dtype": "f32",
                    }
                ),
            ], []),
            "out_of_order_tensor": (
                [
                    TensorDeclaration(
                        {
                            "name": "out_of_order_tensor.A",
                            "num_elements": 1 * 1024 ** 2,
                            "dtype": "f32",
                        }
                    ),
                    TensorDeclaration(
                        {
                            "name": "out_of_order_tensor.B",
                            "num_elements": 2 * 1024 ** 2,
                            "dtype": "f32",
                        }
                    ),
                    TensorDeclaration(
                        {
                            "name": "out_of_order_tensor.C",
                            "num_elements": 3 * 1024 ** 2,
                            "dtype": "f32",
                        }
                    ),
                    TensorDeclaration(
                        {
                            "name": "out_of_order_tensor.D",
                            "num_elements": 4 * 1024 ** 2,
                            "dtype": "f32",
                        }
                    ),
                    TensorDeclaration(
                        {
                            "name": "out_of_order_tensor.E",
                            "num_elements": 5 * 1024 ** 2,
                            "dtype": "f32",
                        }
                    ),
                ],
                [
                    {
                        "trace_id": 0,
                        "action": "tensor_ready",
                        "tensor_name": "out_of_order_tensor.D",
                        "start_time": 0,
                        "end_time": 1,
                    },
                    {
                        "trace_id": 1,
                        "action": "tensor_ready",
                        "tensor_name": "out_of_order_tensor.E",
                        "start_time": 1,
                        "end_time": 2,
                    },
                    {
                        "trace_id": 2,
                        "action": "tensor_ready",
                        "tensor_name": "out_of_order_tensor.A",
                        "start_time": 2,
                        "end_time": 3,
                    },
                    {
                        "trace_id": 3,
                        "action": "tensor_ready",
                        "tensor_name": "out_of_order_tensor.B",
                        "start_time": 3,
                        "end_time": 14,
                    },
                    {
                        "trace_id": 4,
                        "action": "tensor_ready",
                        "tensor_name": "out_of_order_tensor.C",
                        "start_time": 4,
                        "end_time": 5,
                    },
                ],
            ),
        }

        mock_objs = []
        pool = multiprocessing.pool.ThreadPool(nprocs * len(model_dict))
        results = dict([(key, []) for key in model_dict.keys()])
        for i in range(nprocs):
            for (model_name, (tensor_list, spans)) in model_dict.items():
                mock = MockBaguaProcess(
                    i, service_addr, service_port, model_name, tensor_list, spans
                )
                mock_objs.append(mock)
                ret = pool.apply_async(mock.run)
                results[model_name].append(ret)

        pool.close()
        pool.join()
        for root, _, files in os.walk("/tmp", topdown=False):
            for name in files:
                if name.startswith("bagua_autotune_"):
                    autotune_logfile = os.path.join(root, name)
                    print(autotune_logfile)
                    print(open(autotune_logfile).read())

        for ret in results["basic"]:
            hp = ret.get()
            buckets = [[td["name"] for td in bucket] for bucket in hp.buckets]
            self.assertEqual(
                buckets,
                [["basic.A", "basic.B", "basic.C"], ["basic.D"], ["basic.E"]],
                "hp={}".format(hp.dict()),
            )
        for ret in results["Mixed_precision_test"]:
            hp = ret.get()
            buckets = [[td["name"] for td in bucket] for bucket in hp.buckets]
            self.assertEqual(
                buckets,
                [
                    ["Mixed_precision_test.C", "Mixed_precision_test.D"],
                    ["Mixed_precision_test.A", "Mixed_precision_test.B"],
                    ["Mixed_precision_test.E"],
                ],
                "hp={}".format(hp.dict()),
            )
        for ret in results["out_of_order_tensor"]:
            hp = ret.get()
            buckets = [[td["name"] for td in bucket] for bucket in hp.buckets]
            self.assertEqual(
                buckets,
                [
                    ["out_of_order_tensor.D"],
                    ["out_of_order_tensor.E"],
                    [
                        "out_of_order_tensor.A",
                        "out_of_order_tensor.B",
                        "out_of_order_tensor.C",
                    ],
                ],
                "hp={}".format(hp.dict()),
            )

        server.terminate()
        server.join()


if __name__ == "__main__":
    logging.basicConfig(
        level=logging.WARN,
        format="%(filename)s: %(levelname)s: %(funcName)s(): %(lineno)d:\t%(message)s",
    )
    unittest.main()



import logging
import multiprocessing
import bagua_core as B
from bagua.service import AutotuneService
from collections import defaultdict
from . import env
from .env import (
    get_master_addr,
    get_world_size,
    get_rank,
    get_local_rank,
    get_local_size,
    get_default_bucket_size,
    get_bagua_service_port,
)
from enum import IntEnum
from .utils import flatten, unflatten
import torch
import torch.distributed as dist
import torch.distributed.distributed_c10d as c10d
from bagua.service.autotune_service import AutotuneClient
from functools import lru_cache


# must be consistent with Aluminum ReductionOperator: https://github.com/BaguaSys/Aluminum/blob/master/include/aluminum/base.hpp
class ReduceOp(IntEnum):
    """An enum-like class for available reduction operations: SUM, PRODUCT, MIN, MAX, BAND, BOR, BXOR and AVG."""

    SUM = 0
    PRODUCT = 1
    MIN = 2
    MAX = 3
    BOR = 7
    BAND = 8
    BXOR = 9
    AVG = 10


@lru_cache(maxsize=None)
def get_hyperparameters_service_client():
    hyperparameters_service_client = AutotuneClient(
        get_master_addr(), get_bagua_service_port()
    )
    return hyperparameters_service_client


@lru_cache(maxsize=None)
def get_backend(model_name: str):
    backend = B.BaguaCommBackendPy(100, device_id=get_local_rank())
    backend.device_id = get_local_rank()
    backend.stream = torch.cuda.Stream(priority=-1)
    backend.store = c10d._get_default_store()
    backend.internode_communicator = init_bagua_inter_communicator(
        model_name=model_name,
        stream=backend.stream,
        leader_rank=0,
        store=backend.store,
        device_id=backend.device_id,
    )
    backend.intranode_communicator = init_bagua_intra_communicator(
        model_name=model_name,
        stream=backend.stream,
        store=backend.store,
        device_id=backend.device_id,
    )
    backend.global_communicator = init_bagua_communicator(
        model_name=model_name,
        stream=backend.stream,
        store=backend.store,
        device_id=backend.device_id,
    )
    return backend


def run_flask_app():
    from flask import Flask

    autotune_service = AutotuneService(
        world_size=get_world_size(),
        autotune_level=env.get_autotune_level(),
        max_samples=env.get_autotune_max_samples(),
        sampling_confidence_time_s=env.get_autotune_sampling_confidence_time_s(),
        warmup_time_s=env.get_autotune_warmup_time_s(),
        is_output_autotune_log=env.get_is_output_autotune_log(),
        default_bucket_size=get_default_bucket_size(),
    )
    app = Flask(__name__)
    app = autotune_service.setup_app(app)
    log = logging.getLogger("werkzeug")
    log.setLevel(logging.ERROR)

    app.run(host="0.0.0.0", port=get_bagua_service_port())


_autotune_server = None


def start_autotune_server():
    """Start autotune server in background."""
    global _autotune_server

    _autotune_server = multiprocessing.Process(target=run_flask_app)
    _autotune_server.daemon = True
    _autotune_server.start()


def init_process_group():
    """Initializes the PyTorch builtin distributed process group, and this will
    also initialize the distributed package, should be executed before all the
    APIs of bagua.

    Raises:
        RepeatedInitializationError: If you run this function repeatedly

    Examples::
        >>> import bagua.torch_api as bagua
        >>> bagua.init_process_group()
        >>> model = torch.nn.Sequential(
        ...    torch.nn.Linear(D_in, H),
        ...    torch.nn.ReLU(),
        ...    torch.nn.Linear(H, D_out),
        ...    )
        >>> optimizer = torch.optim.SGD(
        ...    model.parameters(),
        ...    lr=0.01,
        ...    momentum=0.9
        ...    )
        >>> model, optimizer = bagua_init(model, optimizer)
    """
    if not dist.is_initialized():
        torch.distributed.init_process_group(
            backend="nccl", init_method="env://"
        )  # fmt: off

    if get_rank() == 0 and _autotune_server is None:
        start_autotune_server()


def gen_nccl_unique_id(comm_type: str, root=0, store=None):
    key = f"{comm_type}-{root}-unique_id"

    if store is None:
        store = c10d._get_default_store()

    if get_rank() == root:
        idstr = B.BaguaSingleCommunicatorPy.generate_nccl_unique_id_str()
        store.set(key, idstr)
    else:
        idstr = store.get(key)
        idstr = str(idstr, encoding="utf-8")

    return idstr


def init_bagua_inter_communicator(
    model_name: str, stream, leader_rank=0, store=None, device_id=None
):
    if device_id is None:
        device_id = get_local_rank()
    nccl_unique_id = gen_nccl_unique_id(
        f"bagua_inter_comm_{model_name}", root=leader_rank, store=store
    )

    if get_rank() % get_local_size() != leader_rank:
        return None

    comm = B.BaguaSingleCommunicatorPy(
        rank=get_rank() // get_local_size(),
        nranks=get_world_size() // get_local_size(),
        device_id=device_id,
        stream_ptr=stream.cuda_stream,
        nccl_unique_id_str=nccl_unique_id,
    )
    comm.cuda_stream = stream
    logging.debug(
        "init bagua internode communicator ok, global rank: %s rank: %s",
        dist.get_rank(),
        comm.rank(),
    )
    return comm


def init_bagua_intra_communicator(model_name: str, stream, store=None, device_id=None):
    if device_id is None:
        device_id = get_local_rank()
    nccl_unique_id = gen_nccl_unique_id(
        f"bagua_intra_comm_{model_name}",
        root=get_rank() // get_local_size() * get_local_size(),
        store=store,
    )

    comm = B.BaguaSingleCommunicatorPy(
        rank=get_rank() % get_local_size(),
        nranks=get_local_size(),
        device_id=device_id,
        stream_ptr=stream.cuda_stream,
        nccl_unique_id_str=nccl_unique_id,
    )
    comm.cuda_stream = stream
    logging.debug(
        "init bagua intranode communicator ok, global rank: %s rank: %s",
        dist.get_rank(),
        comm.rank(),
    )
    return comm


def init_bagua_communicator(model_name: str, stream, store=None, device_id=None):
    if device_id is None:
        device_id = get_local_rank()
    nccl_unique_id = gen_nccl_unique_id(f"bagua_global_comm_{model_name}", store=store)

    comm = B.BaguaSingleCommunicatorPy(
        rank=get_rank(),
        nranks=get_world_size(),
        device_id=device_id,
        stream_ptr=stream.cuda_stream,
        nccl_unique_id_str=nccl_unique_id,
    )
    comm.cuda_stream = stream
    logging.debug(
        "init bagua global communicator ok, global rank: %s rank: %s",
        dist.get_rank(),
        comm.rank(),
    )
    return comm


def send(tensor, dst, comm: B.BaguaSingleCommunicatorPy = None):
    r"""Sends a tensor to dst synchronously.

    Args:
        tensor (torch.Tensor): Data to be sent.
        dst (int): Destination rank.
        comm (B.BaguaSingleCommunicatorPy, optional): The bagua communicator
            to work on. If None, the global bagua communicator will be used.
    """

    assert tensor.device != torch.device("cpu"), "input tensor must be CUDA and dense"

    if comm is None:
        comm = get_backend("").global_communicator

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.send(tensor.to_bagua_tensor().bagua_backend_tensor(), dst)

    torch.cuda.synchronize()


def recv(tensor, src, comm: B.BaguaSingleCommunicatorPy = None):
    r"""Receives a tensor synchronously.

    Args:
        tensor (torch.Tensor): Tensor to fill with received data.
        src (int): Source rank.
        comm (B.BaguaSingleCommunicatorPy, optional): The bagua communicator
            to work on. If None, the global bagua communicator will be used.
    """

    assert tensor.device != torch.device("cpu"), "input tensor must be CUDA and dense"

    if comm is None:
        comm = get_backend("").global_communicator

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.recv(tensor.to_bagua_tensor().bagua_backend_tensor(), src)

    torch.cuda.synchronize()


def broadcast_coalesced(tensors, src=0, comm: B.BaguaSingleCommunicatorPy = None):
    for tensor in tensors:
        assert tensor.device != torch.device(
            "cpu"
        ), "input tensors must be CUDA and dense"

    if comm is None:
        comm = get_backend("").global_communicator

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        coalesced = flatten(tensors)
        comm.broadcast(coalesced.to_bagua_tensor().bagua_backend_tensor(), src)
        for buf, synced in zip(tensors, unflatten(coalesced, tensors)):
            buf.copy_(synced)

    # TODO: remove
    torch.cuda.synchronize()


def broadcast(tensor, src=0, comm: B.BaguaSingleCommunicatorPy = None):
    r"""Broadcasts the tensor to the whole communicator.

    `tensor` must have the same number of elements in all processes
    participating in the collective.

    Args:
        tensor (torch.Tensor): Data to be sent if `root` is the rank of
            current process, and tensor to be used to save received data
            otherwise.
        src (int, optional): Source rank. Defaults to 0.
        comm (B.BaguaSingleCommunicatorPy, optional): The bagua communicator
            to work on. If None, the global bagua communicator will be used.
            Defaults to None.
    """  # noqa: W293

    assert tensor.device != torch.device("cpu"), "input tensor must be CUDA and dense"

    if comm is None:
        comm = get_backend("").global_communicator

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.broadcast(tensor.to_bagua_tensor().bagua_backend_tensor(), src)

    # TODO: remove
    torch.cuda.synchronize()


def reduce(
    send_tensor,
    recv_tensor,
    dst,
    op=ReduceOp.SUM,
    comm: B.BaguaSingleCommunicatorPy = None,
):
    r"""Reduces the tensor across all processes.

    Only the process whit rank `dst` is going to receive the final result.

    Args:
        send_tensor (torch.Tensor): Input of the collective.
        recv_tensor (torch.Tensor): Output of the collective, must have the same size of send_tensor.
        dst (int): Destination rank.
        op (optional): one of the values from `bagua.ReduceOp`
            enum. Specifies an operation used for element-wise reductions.
        comm (B.BaguaSingleCommunicatorPy, optional): The bagua communicator to
            work on. If None the global bagua communicator will be used.
            Defaults to None.
    """  # noqa: W293

    assert send_tensor.device != torch.device(
        "cpu"
    ), "send tensor must be CUDA and dense"
    assert recv_tensor.device != torch.device(
        "cpu"
    ), "recv tensor must be CUDA and dense"

    if comm is None:
        comm = get_backend("").global_communicator

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.reduce(
            send_tensor.to_bagua_tensor().bagua_backend_tensor(),
            recv_tensor.to_bagua_tensor().bagua_backend_tensor(),
            dst,
            int(op),
        )

    torch.cuda.synchronize()


def reduce_inplace(
    tensor, dst, op=ReduceOp.SUM, comm: B.BaguaSingleCommunicatorPy = None
):
    r"""The inplace version of reduce."""

    assert tensor.device != torch.device("cpu"), "input tensor must be CUDA and dense"

    if comm is None:
        comm = get_backend("").global_communicator

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.reduce_inplace(
            tensor.to_bagua_tensor().bagua_backend_tensor(), dst, int(op)
        )

    torch.cuda.synchronize()


def allreduce_coalesced_inplace(
    tensors,
    op=ReduceOp.SUM,
    comm: B.BaguaSingleCommunicatorPy = None,
):
    for tensor in tensors:
        assert tensor.device != torch.device(
            "cpu"
        ), "input tensors must be CUDA and dense"

    if comm is None:
        comm = get_backend("").global_communicator

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        coalesced = flatten(tensors)
        comm.allreduce_inplace(
            coalesced.to_bagua_tensor("allreduce_coalesced"), int(op)
        )

        for buf, synced in zip(tensors, unflatten(coalesced, tensors)):
            buf.copy_(synced)

    # TODO: remove
    torch.cuda.synchronize()


def allreduce(
    send_tensor,
    recv_tensor,
    op=ReduceOp.SUM,
    comm: B.BaguaSingleCommunicatorPy = None,
):
    """Reduces the tensor data across all machines in such a way that all get
    the final result. After the call recv_tensor is going to be bitwise identical
    in all processes.

    Args:
        send_tensor (torch.Tensor): Input of the collective.
        recv_tensor (torch.Tensor): Output of the collective, must have the same size of send_tensor.
        op (optional): one of the values from `bagua.ReduceOp` enum. Specifies an operation used for element-wise reductions.
        comm (B.BaguaSingleCommunicatorPy, optional): The bagua communicator to
            work on. If None the global bagua communicator will be used.
            Defaults to None.

    Examples:
        >>> from bagua.torch_api import allreduce
        >>> # All tensors below are of torch.int64 type.
        >>> # We have 2 process groups, 2 ranks.
        >>> send_tensor = torch.arange(2, dtype=torch.int64) + 1 + 2 * rank
        >>> recv_tensor = torch.zeros(2, dtype=torch.int64)
        >>> send_tensor
        tensor([1, 2]) # Rank 0
        tensor([3, 4]) # Rank 1
        >>> allreduce(send_tensor, recv_tensor)
        >>> recv_tensor
        tensor([4, 6]) # Rank 0
        tensor([4, 6]) # Rank 1

        >>> # All tensors below are of torch.cfloat type.
        >>> # We have 2 process groups, 2 ranks.
        >>> send_tensor = torch.tensor([1+1j, 2+2j], dtype=torch.cfloat) + 2 * rank * (1+1j)
        >>> recv_tensor = torch.zeros(2, dtype=torch.cfloat)
        >>> send_tensor
        tensor([1.+1.j, 2.+2.j]) # Rank 0
        tensor([3.+3.j, 4.+4.j]) # Rank 1
        >>> allreduce(send_tensor, recv_tensor)
        >>> recv_tensor
        tensor([4.+4.j, 6.+6.j]) # Rank 0
        tensor([4.+4.j, 6.+6.j]) # Rank 1
    """  # noqa: E501

    assert send_tensor.device != torch.device(
        "cpu"
    ), "send tensor must be CUDA and dense"
    assert recv_tensor.device != torch.device(
        "cpu"
    ), "recv tensor must be CUDA and dense"

    if comm is None:
        comm = get_backend("").global_communicator

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.allreduce(
            send_tensor.to_bagua_tensor().bagua_backend_tensor(),
            recv_tensor.to_bagua_tensor().bagua_backend_tensor(),
            int(op),
        )

    # TODO: remove
    torch.cuda.synchronize()


def allreduce_inplace(
    tensor,
    op=ReduceOp.SUM,
    comm: B.BaguaSingleCommunicatorPy = None,
):
    """The inplace version of allreduce."""

    assert tensor.device != torch.device("cpu"), "input tensor must be CUDA and dense"

    if comm is None:
        comm = get_backend("").global_communicator

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.allreduce_inplace(tensor.to_bagua_tensor().bagua_backend_tensor(), int(op))

    torch.cuda.synchronize()


def allgather(
    send_tensor,
    recv_tensor,
    comm: B.BaguaSingleCommunicatorPy = None,
):
    """Gathers send_tensors from all machines to recv_tensor.

    Args:
        send_tensor (torch.Tensor): Input of the collective.
        recv_tensor (torch.Tensor): Output of the collective, must have size send_tensor.size()*comm.nranks.
        comm (B.BaguaSingleCommunicatorPy, optional): The bagua communicator to
            work on. If None the global bagua communicator will be used.
            Defaults to None.
    """

    assert send_tensor.device != torch.device(
        "cpu"
    ), "send tensor must be CUDA and dense"
    assert recv_tensor.device != torch.device(
        "cpu"
    ), "recv tensor must be CUDA and dense"

    if comm is None:
        comm = get_backend("").global_communicator

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.allgather(
            send_tensor.to_bagua_tensor().bagua_backend_tensor(),
            recv_tensor.to_bagua_tensor().bagua_backend_tensor(),
        )

    torch.cuda.synchronize()


def allgather_inplace(
    tensor,
    comm: B.BaguaSingleCommunicatorPy = None,
):
    """The inplace version of allgather."""

    assert tensor.device != torch.device("cpu"), "input tensor must be CUDA and dense"

    if comm is None:
        comm = get_backend("").global_communicator

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.allgather_inplace(tensor.to_bagua_tensor().bagua_backend_tensor())

    torch.cuda.synchronize()


def gather(
    send_tensor,
    recv_tensor,
    dst,
    comm: B.BaguaSingleCommunicatorPy = None,
):
    """Gathers send_tensors from all machines to recv_tensor in a single process.

    Args:
        send_tensor (torch.Tensor): Input of the collective.
        recv_tensor (torch.Tensor): Output of the collective, must have size send_tensor.size()*comm.nranks.
        dst (int): Destination rank.
        comm (B.BaguaSingleCommunicatorPy, optional): The bagua communicator to
            work on. If None the global bagua communicator will be used.
            Defaults to None.
    """

    assert send_tensor.device != torch.device(
        "cpu"
    ), "send tensor must be CUDA and dense"
    assert recv_tensor.device != torch.device(
        "cpu"
    ), "recv tensor must be CUDA and dense"

    if comm is None:
        comm = get_backend("").global_communicator

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.gather(
            send_tensor.to_bagua_tensor().bagua_backend_tensor(),
            recv_tensor.to_bagua_tensor().bagua_backend_tensor(),
            dst,
        )

    torch.cuda.synchronize()


def gather_inplace(
    tensor,
    count,
    dst,
    comm: B.BaguaSingleCommunicatorPy = None,
):
    """The inplace version of gather.

    Args:
        tensor (torch.Tensor): Input and output of the collective, For dst process,
            has size count*comm.nranks() and acts as recv_tensor above. For non-dst processes,
            has size count and acts as send_tensor above.
        count (int): The per-rank data count.
        dst (int): Destination rank.
        comm (B.BaguaSingleCommunicatorPy, optional): The bagua communicator to
            work on. If None the global bagua communicator will be used.
            Defaults to None.
    """

    assert tensor.device != torch.device("cpu"), "input tensor must be CUDA and dense"

    if comm is None:
        comm = get_backend("").global_communicator

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.gather_inplace(tensor.to_bagua_tensor().bagua_backend_tensor(), count, dst)

    torch.cuda.synchronize()


def scatter(
    send_tensor,
    recv_tensor,
    src,
    comm: B.BaguaSingleCommunicatorPy = None,
):
    """Scatters send_tensor to all machines.

    Args:
        send_tensor (torch.Tensor): Input of the collective, must have size recv_tensor.size()*comm.nranks.
        recv_tensor (torch.Tensor): Output of the collective.
        src (int): Source rank.
        comm (B.BaguaSingleCommunicatorPy, optional): The bagua communicator to
            work on. If None the global bagua communicator will be used.
            Defaults to None.
    """

    assert send_tensor.device != torch.device(
        "cpu"
    ), "send tensor must be CUDA and dense"
    assert recv_tensor.device != torch.device(
        "cpu"
    ), "recv tensor must be CUDA and dense"

    if comm is None:
        comm = get_backend("").global_communicator

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.scatter(
            send_tensor.to_bagua_tensor().bagua_backend_tensor(),
            recv_tensor.to_bagua_tensor().bagua_backend_tensor(),
            src,
        )

    torch.cuda.synchronize()


def scatter_inplace(
    tensor,
    count,
    src,
    comm: B.BaguaSingleCommunicatorPy = None,
):
    """The inplace version of gather.

    Args:
        tensor (torch.Tensor): Input and output of the collective, For src process,
            has size count*comm.nranks() and acts as send_tensor above. For non-src processes,
            has size count and acts as recv_tensor above.
        count (int): The per-rank data count.
        src (int): Source rank.
        comm (B.BaguaSingleCommunicatorPy, optional): The bagua communicator to
            work on. If None the global bagua communicator will be used.
            Defaults to None.
    """

    assert tensor.device != torch.device("cpu"), "input tensor must be CUDA and dense"

    if comm is None:
        comm = get_backend("").global_communicator

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.scatter_inplace(
            tensor.to_bagua_tensor().bagua_backend_tensor(), count, src
        )

    torch.cuda.synchronize()


def reduce_scatter(
    send_tensor,
    recv_tensor,
    op=ReduceOp.SUM,
    comm: B.BaguaSingleCommunicatorPy = None,
):
    """Reduces on send_tensor, then scatters send_tensor to all machines.

    Args:
        send_tensor (torch.Tensor): Input of the collective, must have size recv_tensor.size()*comm.nranks.
        recv_tensor (torch.Tensor): Output of the collective.
        op (optional): one of the values from `bagua.ReduceOp` enum. Specifies an operation used for element-wise reductions.
        comm (B.BaguaSingleCommunicatorPy, optional): The bagua communicator to
            work on. If None the global bagua communicator will be used.
            Defaults to None.
    """

    assert send_tensor.device != torch.device(
        "cpu"
    ), "send tensor must be CUDA and dense"
    assert recv_tensor.device != torch.device(
        "cpu"
    ), "recv tensor must be CUDA and dense"

    if comm is None:
        comm = get_backend("").global_communicator

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.reduce_scatter(
            send_tensor.to_bagua_tensor().bagua_backend_tensor(),
            recv_tensor.to_bagua_tensor().bagua_backend_tensor(),
            int(op),
        )

    torch.cuda.synchronize()


def reduce_scatter_inplace(
    tensor,
    op=ReduceOp.SUM,
    comm: B.BaguaSingleCommunicatorPy = None,
):
    """The inplace version of reduce_scatter.

    Args:
        send_tensor (torch.Tensor): Input and output of the collective, must satisfy: `tensor.size() % comm.nranks == 0`.
        op (optional): one of the values from `bagua.ReduceOp` enum. Specifies an operation used for element-wise reductions.
        comm (B.BaguaSingleCommunicatorPy, optional): The bagua communicator to
            work on. If None the global bagua communicator will be used.
            Defaults to None.
    """

    assert tensor.device != torch.device("cpu"), "send tensor must be CUDA and dense"

    if comm is None:
        comm = get_backend("").global_communicator

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.reduce_scatter_inplace(
            tensor.to_bagua_tensor().bagua_backend_tensor(), int(op)
        )

    torch.cuda.synchronize()


def alltoall(
    send_tensor,
    recv_tensor,
    comm: B.BaguaSingleCommunicatorPy = None,
):
    """All processes send data to all processes.

    Args:
        send_tensor (torch.Tensor): Input of the collective, must satisfy: `send_tensor.size() % comm.nranks == 0`.
        recv_tensor (torch.Tensor): Output of the collective, must have the same size of send_tensor.
        comm (B.BaguaSingleCommunicatorPy, optional): The bagua communicator to
            work on. If None the global bagua communicator will be used.
            Defaults to None.
    """

    assert send_tensor.device != torch.device(
        "cpu"
    ), "send tensor must be CUDA and dense"
    assert recv_tensor.device != torch.device(
        "cpu"
    ), "recv tensor must be CUDA and dense"

    if comm is None:
        comm = get_backend("").global_communicator

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.alltoall(
            send_tensor.to_bagua_tensor().bagua_backend_tensor(),
            recv_tensor.to_bagua_tensor().bagua_backend_tensor(),
        )

    torch.cuda.synchronize()


def alltoall_inplace(
    tensor,
    comm: B.BaguaSingleCommunicatorPy = None,
):
    """The inplace version of alltoall."""

    assert tensor.device != torch.device("cpu"), "recv tensor must be CUDA and dense"

    if comm is None:
        comm = get_backend("").global_communicator

    event = torch.cuda.current_stream().record_event()
    comm.cuda_stream.wait_event(event)

    with torch.cuda.stream(comm.cuda_stream):
        comm.alltoall_inplace(tensor.to_bagua_tensor().bagua_backend_tensor())

    torch.cuda.synchronize()



import bagua.torch_api as bagua
import torch
import torch.optim as optim
import unittest
import os
from tests.internal.common_utils import find_free_port


def run_step(opt, flag_param, fuse, wrap, device):
    weight = torch.tensor(
        [[-0.2109, -0.4976], [-0.1413, -0.3420], [-0.2524, 0.6976]],
        requires_grad=True,
    )
    bias = torch.tensor(
        [-0.1085, -0.2979, 0.6892],
        requires_grad=True,
    )
    weight2 = torch.tensor(
        [[-0.0508, -0.3941, -0.2843]],
        requires_grad=True,
    )
    bias2 = torch.tensor([-0.0711], requires_grad=True)

    model = torch.nn.Sequential(
        torch.nn.Linear(2, 3),
        torch.nn.Sigmoid(),
        torch.nn.Linear(3, 1),
        torch.nn.Sigmoid(),
    )

    pretrained_dict = model.state_dict()
    pretrained_dict["0.weight"] = weight
    pretrained_dict["0.bias"] = bias
    pretrained_dict["2.weight"] = weight2
    pretrained_dict["2.bias"] = bias2
    model.load_state_dict(pretrained_dict)

    model = model.to(device)
    optimizer = opt(model.parameters(), **flag_param)

    if fuse:
        bagua.contrib.FusedOptimizer(optimizer, do_flatten=not wrap)

    if wrap:
        model.with_bagua(
            [optimizer],
            bagua.algorithms.gradient_allreduce.GradientAllReduceAlgorithm(),
        )

    input = torch.tensor([0.1, 0.2, 0.3, 0.4, 0.5, 0.6], device=device).reshape(3, 2)

    for _ in range(1001):
        optimizer.zero_grad()
        output = model(input)
        loss = output.sum()
        loss.backward()

        optimizer.step()

    return model.parameters()


class TestFusedOptimizer(unittest.TestCase):
    def run_all_optimizers_once(self, wrap, device):
        optimizer_list = [
            optim.Adam,
            optim.Adam,
            optim.Adam,
            optim.Adam,
            optim.AdamW,
            optim.AdamW,
            optim.AdamW,
            optim.AdamW,
            optim.SGD,
            optim.SGD,
            optim.RMSprop,
            optim.RMSprop,
            optim.RMSprop,
            optim.RMSprop,
            optim.Rprop,
            optim.ASGD,
            optim.ASGD,
            optim.Adamax,
            optim.Adamax,
            optim.Adadelta,
            optim.Adadelta,
        ]

        flag_params = [
            dict(weight_decay=1.0, amsgrad=True),  # Adam
            dict(weight_decay=1.0, amsgrad=False),  # Adam
            dict(weight_decay=0.0, amsgrad=True),  # Adam
            dict(weight_decay=0.0, amsgrad=False),  # Adam
            dict(weight_decay=1.0, amsgrad=True),  # AdamW
            dict(weight_decay=1.0, amsgrad=False),  # AdamW
            dict(weight_decay=0.0, amsgrad=True),  # AdamW
            dict(weight_decay=0.0, amsgrad=False),  # AdamW
            dict(lr=0.2, momentum=1, dampening=0, weight_decay=1, nesterov=True),  # SGD
            dict(
                lr=0.2, momentum=1, dampening=0.5, weight_decay=1, nesterov=False
            ),  # SGD
            dict(weight_decay=1, momentum=1, centered=True),  # RMSprop
            dict(weight_decay=1, momentum=0, centered=True),  # RMSprop
            dict(weight_decay=1, momentum=1, centered=False),  # RMSprop
            dict(weight_decay=0, momentum=1, centered=False),  # RMSprop
            dict(lr=1e-2, etas=(0.5, 1.2), step_sizes=(1e-6, 50)),  # Rprop
            dict(weight_decay=0),  # ASGD
            dict(weight_decay=1),  # ASGD
            dict(weight_decay=0),  # Adamax
            dict(weight_decay=1),  # Adamax
            dict(weight_decay=0),  # Adadelta
            dict(weight_decay=1),  # Adadelta
        ]

        for opt, flag_param in zip(optimizer_list, flag_params):
            res1 = run_step(opt, flag_param, fuse=True, wrap=wrap, device=device)
            res2 = run_step(opt, flag_param, fuse=False, wrap=wrap, device=device)

            for p1, p2 in zip(res1, res2):
                self.assertTrue(torch.equal(p1, p2))

    def test_fused_optimizer(self):
        self.run_all_optimizers_once(device="cpu", wrap=False)

    def test_fused_optimizer_with_bagua_wrapper(self):
        if not torch.cuda.is_available():
            print("skip tests since cuda is not available")
            return

        # init env
        os.environ["WORLD_SIZE"] = "1"
        os.environ["LOCAL_WORLD_SIZE"] = "1"
        os.environ["MASTER_ADDR"] = "127.0.0.1"
        os.environ["MASTER_PORT"] = str(find_free_port())
        os.environ["BAGUA_SERVICE_PORT"] = str(find_free_port())

        os.environ["RANK"] = "0"
        os.environ["LOCAL_RANK"] = "0"

        # init bagua distributed process group
        torch.cuda.set_device(bagua.get_local_rank())
        bagua.init_process_group()

        self.run_all_optimizers_once(device="cuda:0", wrap=True)


if __name__ == "__main__":
    unittest.main()



import socket


def find_free_port():
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind(("localhost", 0))
    sockname = sock.getsockname()
    sock.close()
    return sockname[1]



import unittest
import torch
from torch.utils.data import TensorDataset, DataLoader
from bagua.torch_api.contrib import (
    LoadBalancingDistributedSampler,
    LoadBalancingDistributedBatchSampler,
)


class TestLoadBalancingDataLoader(unittest.TestCase):
    def test_load_balancing_distributed_sampler(self):
        return
        n = 10
        dataset = TensorDataset(torch.randn(n, 2), torch.randperm(n))
        sampler = LoadBalancingDistributedSampler(
            dataset, complexity_fn=lambda x: x[1], num_replicas=1, rank=0, shuffle=False
        )

        dataloader = DataLoader(
            dataset,
            batch_size=1,
            shuffle=sampler is None,
            sampler=sampler,
        )

        for i, data in enumerate(dataloader):
            self.assertTrue(i == data[1].item())

    def test_load_balancing_distributed_batch_sampler(self):
        num_replicas = 1
        total_batch = 5

        n = sum([i + 1 for i in range(total_batch)]) * num_replicas
        dataset = TensorDataset(torch.randn(n, 2), torch.randperm(n))

        def batch_fn(indices):
            batch = []
            batch_size = 1
            for i in indices:
                batch.append(i)
                if len(batch) == batch_size:
                    yield batch
                    batch = []
                    batch_size += 1
            if len(batch) > 0:
                yield batch

        sampler = LoadBalancingDistributedSampler(
            dataset,
            complexity_fn=lambda x: x[1],
            num_replicas=num_replicas,
            rank=0,
            shuffle=False,
        )

        batch_sampler = LoadBalancingDistributedBatchSampler(
            sampler,
            batch_fn=lambda x: list(batch_fn(x)),
        )

        dataloader = torch.utils.data.DataLoader(dataset, batch_sampler=batch_sampler)

        cur_idx = 0
        for i, data in enumerate(dataloader):
            batch_size = data[0].shape[0]
            self.assertTrue(batch_size == i + 1)
            self.assertTrue(cur_idx == data[1][0].item())

            cur_idx += batch_size * num_replicas


if __name__ == "__main__":
    unittest.main()
